package xcsp2ep;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import xcsp2ep.mapper.Mapper;
import xcsp2ep.mapper.MapperException;
import xcsp2ep.parser.InstanceParser;
import xcsp2ep.parser.XCSPInstance;
import xcsp2ep.mapper.EssencePModel;

public class Translation{

	// the file we will write the Essence' instance into
	public static final String OUTPUT_FILE_NAME = "out.eprime";
	public static final String ESSENCE_PRIME_HEADER = "language ESSENCE' 1.b.a\n";
	
	public Translation() {
	}
	
	/**
	 * Translate the XCSP instance in 'inputFile' into Essence'
	 * and write it into file 'outputFile'
	 * 
	 * @param inputFile
	 * @param outputFile
	 * @throws MapperException
	 * @throws Exception
	 */
	public static void translate(String inputFile, String outputFile) 
		throws MapperException, Exception {
		
		writeStartMessage(inputFile);
		long startTime = System.currentTimeMillis();
		
		// parse XCSP instance
		boolean displayParsedInstance = false;
		InstanceParser parser = new InstanceParser();
		parser.loadInstance(inputFile);
		XCSPInstance xcspInstance = parser.parse(displayParsedInstance);
			
		// map the XCSP syntax tree to Essence
		Mapper mapper = new Mapper();
		EssencePModel essencePmodel = mapper.mapToEssencePrime(xcspInstance);
		writeStringIntoFile(outputFile, essencePmodel.toString());
		long stopTime = System.currentTimeMillis();
		
		double translationTime = (stopTime - startTime) / 1000.0;
		
		writeEndMessage(outputFile, translationTime);
	}
	
	
	public static String translate(String fileName) 
		throws MapperException, Exception {
		
		writeStartMessage(fileName);
		long startTime = System.currentTimeMillis();
		
		// parse the input file
		boolean displayParsedInstance = false;
		InstanceParser parser = new InstanceParser();
		parser.loadInstance(fileName);
		XCSPInstance xcspInstance = parser.parse(displayParsedInstance);
			
		// map the XCSP syntax tree to Essence
		Mapper mapper = new Mapper();
		EssencePModel essencePmodel = mapper.mapToEssencePrime(xcspInstance);
		writeStringIntoFile(OUTPUT_FILE_NAME, essencePmodel.toString());
		long stopTime = System.currentTimeMillis();
		
		double translationTime = (stopTime - startTime) / 1000.0;
		
		writeEndMessage(OUTPUT_FILE_NAME, translationTime);
		return essencePmodel.toString();
	}
	
	
	private static void writeStringIntoFile(String fileName, String stringToWrite) 
	throws IOException {
	
		File file = new File(fileName);
		file.createNewFile();
	
		FileWriter writer = new FileWriter(file);
		writer.write(stringToWrite);
    	
		writer.flush();
		writer.close();
	
	}

	public static void writeStartMessage(String inputFileName) {
		System.out.println("# Welcome to xcsp2ep v0.1"+
				"\n# A converter from Format XCSP 2.1 to Essence' 1.b.a"+
				"\n# Bug reports: andrea@cs.st-and.ac.uk"
				+"\n\nConverting XCSP file '"+inputFileName+"' to Essence' ...");
	}
	
	public static void writeEndMessage(String outputFileName, double translationTime) {
		System.out.println("Conversion successfull.\nWritten Essence' instance into file '"+outputFileName+"'.\n" +
				"Translation time: "+translationTime+" sec\n");
	}
	
	public static void writeHelpMessage() {
		
		System.out.println("Usage: xcsp2ep inputFile [outputFile] "+
				"\n\ninputfile:\tfile that contains XCSP instance"+
				"\noutputfile:\tfile where Essence' instance should be written into (optional)"+
				"\n\t\tdefault output file:"+OUTPUT_FILE_NAME+
				"\nbug reports: andrea@cs.st-and.ac.uk\n");
	}
	
	public static void main(String[] args) {
		
		try {
			
			if(args.length == 1) 
				translate(args[0]);
			
			else if(args.length == 2)
				translate(args[0], args[1]);
			
			else writeHelpMessage();
			
		} catch (Exception e) {
			e.printStackTrace(System.out);
			System.out.println("Cannot convert XCSP instance to Essence' instance.\n");
			writeHelpMessage();
			System.exit(1);
		}
		
	}
}
