package xcsp2ep.mapper;

import java.util.HashMap;
import java.util.ArrayList;

import xcsp2ep.expression.Domain;
import xcsp2ep.expression.BoolDomain;
import xcsp2ep.expression.Expression;
import xcsp2ep.expression.Variable;
import xcsp2ep.expression.ConstantArray;
import xcsp2ep.mapper.Objective;

public class EssencePModel {

	public final String ESSENCE_P_VERSION = "language ESSENCE' 1.b.a";
	
	/** parameter values that we might not have inserted yet
	because they occur in quantifications */
	public HashMap<String, ConstantArray> constantArrays;
	HashMap<String, int[]> constantOffsetsFromZero;
	
	/** decision variables and their corresponding domain */
	HashMap<String, Domain> decisionVariables;
	ArrayList<String> decisionVariablesNames;
	
	/** constraints list */
	ArrayList<Expression> constraintList;
	
	/** the objective expression */
	Objective objective;
	
	/** auxiliary variables */
	ArrayList<Variable> auxiliaryVariables;
	
	int usedCommonSubExpressions;
	int usedEqualSubExpressions;
	
	//=============== CONSTRUCTORS ==================================
	
	public EssencePModel(HashMap<String, Domain> decisionVariables,
            ArrayList<String> decisionVariablesNames,
            ArrayList<Expression> constraints,
            Objective objective) {

		this.decisionVariables = decisionVariables;
		this.decisionVariablesNames = decisionVariablesNames;
		this.constraintList = constraints;
		this.objective = objective;
		this.auxiliaryVariables = new ArrayList<Variable>();
		this.usedCommonSubExpressions = 0;
		this.usedEqualSubExpressions = 0;
		this.constantArrays = new HashMap<String, ConstantArray>();
		this.constantOffsetsFromZero = new HashMap<String,int[]>();
	}
	
	public EssencePModel(HashMap<String, Domain> decisionVariables,
			               ArrayList<String> decisionVariablesNames,
			               ArrayList<Expression> constraints,
			               HashMap<String, ConstantArray> constantArrays,
			               HashMap<String,int[]> constantArrayOffsets,
			               Objective objective) {
		
		this.decisionVariables = decisionVariables;
		this.decisionVariablesNames = decisionVariablesNames;
		this.constraintList = constraints;
		this.constantArrays = constantArrays;
		this.constantOffsetsFromZero = constantArrayOffsets;
		
		this.objective = objective;
		this.auxiliaryVariables = new ArrayList<Variable>();
		this.usedCommonSubExpressions = 0;
	}
	 
	// =============== METHODS =======================================


	public int getOffsetFromZeroAt(String arrayName, int index) 
		throws MapperException {
		
		int offsets[] = this.constantOffsetsFromZero.get(arrayName);
		
		// the arrayName is not a 
		if(offsets == null)
			throw new MapperException("Trying to get offset from zero of unknown constant array '"+
					arrayName+"' at index:"+index);
		
		if(index <0 || index >= offsets.length)
			throw new MapperException("Trying to get offset from zero of constant array '"+
					arrayName+"' at index out of bounds :"+index);
		
		return offsets[index];
	}
	
	
	
	

	public Variable getLastAddedAuxiliaryVariable() {
		if(this.auxiliaryVariables.size() ==0)
			return null;
		return this.auxiliaryVariables.get(this.auxiliaryVariables.size()-1);
	}
	
	public void removeLastAuxiliaryVariable() {
		if(this.auxiliaryVariables.size() > 0)
			this.auxiliaryVariables.remove(this.auxiliaryVariables.size()-1);
	}
	
	public int getAmountOfCommonSubExpressionsUsed() {
		return this.usedCommonSubExpressions;
	}
	
	public int getAmountOfEqualSubExpressionsUsed() {
		return this.usedEqualSubExpressions;
	}
	

	
	/**
	 * Given the String representation of the original variable, we 
	 * return the equal variable, with which it is going to be replaced.
	 * 
	 * @param originalVariable
	 * @return
	 */
	
	
	public void setAmountOfCommonSubExpressionsUsed(int noCommonSubExpressionsUsed) {
		this.usedCommonSubExpressions = noCommonSubExpressionsUsed;
	}
	
	public void setAmountOfEqualSubExpressionsUsed(int number) {
		this.usedEqualSubExpressions = number;
	}
	
	public void setObjectiveExpression(Expression objExpression) {
		if(this.objective.objective != null)
			this.objective.objective = objExpression;
	}
	
	
	/**
	 * Returns true, if there exists a decision variable with 
	 * the name variableName
	 */
	public boolean isDecisionVariable(String variableName) {
		return (this.decisionVariables.get(variableName) != null);
	}
	
	/**
	 * 
	 * @param variableName
	 * @return the domain of the variable with variableName. If there
	 * 	exists no variable with that name it returns null.  
	 */
	public Domain getDomainOfVariable(String variableName) {
		if(this.decisionVariables.containsKey(variableName))
			return this.decisionVariables.get(variableName);
		else return null;
	}
	
	public Expression getObjectiveExpression() {
		return this.objective.objective;
	}
	
	public boolean isObjectiveMaximising() {
		return this.objective.maximise;
	}
	
	public void setFlattenedObjectiveExpression(Expression objective) {
		this.objective.objective = objective;
	}
	
	public ArrayList<String> getDecisionVariablesNames() {
		return this.decisionVariablesNames;
	}
	
	public ArrayList<Expression> getConstraints() {
		return this.constraintList;
	}
	
	
	public void replaceConstraintsWith(ArrayList<Expression> newConstraints) {
		this.constraintList.clear();
		this.constraintList = newConstraints;
	}
	
	public void addAuxiliaryVariable(Variable variable) {
		this.auxiliaryVariables.add(variable);
	}
	
	public ArrayList<Variable> getAuxVariables() {
		return this.auxiliaryVariables;
	}
	
	public String toString() {
		
		// header
		StringBuffer s = new StringBuffer(this.ESSENCE_P_VERSION+"\n\n");
		
	
		// decision variables
		ArrayList<String> booleanVariables = new ArrayList<String>();
		for(int i=0; i<this.decisionVariablesNames.size(); i++) {
			String variableName = decisionVariablesNames.get(i);
			if(this.decisionVariables.get(variableName) instanceof BoolDomain)
				booleanVariables.add(variableName);
			else s.append("find\t"+variableName+"\t: "+this.decisionVariables.get(variableName)+"\n");
		}
		if(booleanVariables.size() > 0) {
			s.append("find\t"+booleanVariables.get(0));
			for(int i=1; i<booleanVariables.size(); i++) {
				if(i % 20 == 0) 
					s.append(",\n\t\t"+booleanVariables.get(i));
				else 
					s.append(", "+booleanVariables.get(i));
			}
			s.append(" : bool\n");
		}
		
		s.append("\n");
		// auxiliary variables
		for(int i=0; i<this.auxiliaryVariables.size(); i++) {
			Variable auxVar = auxiliaryVariables.get(i);
			if(i%5 ==0) s.append("\n$");
			s.append("  "+auxVar+" : {"+auxVar.getDomain()[0]+", "+auxVar.getDomain()[1]+"}  ");
		}
		s.append(printStatistics()+"\n");
		
		
		// objective
		if(!this.objective.isEmptyObjective())
			s.append("\n"+this.objective.toString()+"\n");
		
		// constraints
		s.append("such that\n");
		
		if(this.constraintList.size() == 0)
			return s.toString();
		
		for(int i=0; i<this.constraintList.size()-1; i++)
			s.append("\t"+constraintList.get(i)+",\n\n");
		s.append("\t"+constraintList.get(constraintList.size()-1)+"\n");
		
		return s.toString();
	}
	
	
	private String printStatistics() {
		String s = "\n$ Statistical data about translation\n";
		s = s.concat("$ amount of subexpressions used: "+this.usedCommonSubExpressions+"\n");
		s = s.concat("$ amount of nested subexpressions used: "+this.usedEqualSubExpressions+"\n"); 
		return s;
	}
	
	
	public HashMap<String, Domain> getDecisionVariables() {
		return this.decisionVariables;
	}
}
